class Ball extends GameEntity {
  float radius = 5;
  PVector position = new PVector(0f, 0f);
  PVector velocity = new PVector(0f, 0f);
  boolean stickToPaddle = false;
  
  color fillColor = color(150, 240, 255);
  
  Ball(float x, float y){
    position.set(x, y, 0);
  }
  
  Ball(float x, float y, float radius){
    this(x, y);
    this.radius = radius;
  }
  
  Ball(float x, float y, float radius, boolean stick){
    this(x, y, radius);
    this.stickToPaddle = stick;
  }
  
  void update(float deltaTime){
    constrainVelocity();
    
    if ( stickToPaddle ){
      //Stay on top of the paddle, smack in the middle
      position.x = state.paddle.position.x + state.paddle.width / 2;
      position.y = state.paddle.position.y - radius - 1;
    } else {
      /*
      Perform continuous collision detection!
      
      To do this, we calculate the path the ball has travelled in the previous frame 
      and find the closest point where it intersects the bounding box of another game
      entity. See intersectLineWithExpandedAABB() in "Collision" for details.
      
      If a collision is detected, we reflect the velocity vector around the surface 
      normal of the box edge we hit, set the ball's position to the collision point and 
      restart the process from the beginning. 
      */
      
      int collisionCount = 0;
      PLine path = new PLine();
      PVector endPoint = new PVector();
      
      while ( (collisionCount < 5) && (deltaTime > 0) ){
        //Calculate the path the ball has travelled during the elapsed time
        endPoint.set(velocity);
        endPoint.mult(deltaTime);
        endPoint.add(position);
        path.set(position, endPoint);
        //Check if the path collides with anything in the environment
        Collision closestCollision = null;
        for(GameEntity entity: state.entities){
          Collision collision = entity.checkForCollision(path, radius);  
          if ( (collision != null) && ( (closestCollision == null) || (collision.t < closestCollision.t) )){
            closestCollision = collision;
            closestCollision.target = entity;
          }
        }
        
        if ( closestCollision != null ){
          collisionCount++;
          velocity = bounceVector(velocity, closestCollision.surfaceNormal);
          velocity.mult(closestCollision.target.restitution); //Apply the "bounciness" factor of the thing we hit
          constrainVelocity();
          position.set(closestCollision.position);
          
          //Notify the object we collided with
          closestCollision.target.onCollision(closestCollision, this); //Note: may alter the ball's properties
          
          //Calculate how much time we "used up" travelling to the collision point and 
          //decrease the delta accordingly.
          deltaTime -= (deltaTime * closestCollision.t); //
        } else {
          position.set(path.end);
          break;
        }
      }
    }
    
    //Check if the ball has left the arena
    float threshold = 5*radius; 
    Arena arena = state.arena;
    if ( 
      ( -position.x > threshold) || 
      ((position.x - arena.width) > threshold) ||
      ( -position.y > threshold) ||
      ((position.y - arena.height) > threshold)
    ){
      die();
    }
  }
  
  void constrainVelocity(){
    float magnitude = velocity.mag();
    if ( magnitude > maxBallVelocity ){
      velocity.normalize();
      velocity.mult(maxBallVelocity);
    } else if ( magnitude < minBallVelocity ){
      velocity.normalize();
      velocity.mult(minBallVelocity);
    }
  }
  
  void draw(){
    fill(fillColor);
    stroke(blendColor(fillColor, color(100), SUBTRACT)); //Compute a darker version of the fill color
    strokeWeight(2);
    ellipse(position.x, position.y, radius*2, radius*2);
  }
  
  void die(){
    dead = true;
    state.ballLost();
  }
}
